/**
\brief      This module implements the gnocl::sourceView widget.
\authors    Peter G. Baum, William J Giddings
\date       2001-03:
*/

/* user documentation */

/**
  \page page53 gnocl::sourceView
  \section sec gnocl::sourceView
  Implementation of gnocl::sourceView
  \subsection subsection1 Implemented Options
  \verbatim

  \endverbatim
  \subsection subsection2 Implemented Commands
  \verbatim

  \endverbatim
  \subsection subsection3 Sample Tcl Script
  \code
   set box [gnocl::box -buttonType 1]
   $box add [gnocl::button -text "%_Do_n't save" -icon "%#No"]
   $box add [gnocl::button -text "%#Cancel"]
   $box add [gnocl::button -text "%#Save"]
   gnocl::window -title "Buttons" -child $box
  \endcode
  \subsection subsection4 Produces
  \image html "../pics/sourceView.png"
*/
/*
http://library.gnome.org/devel/gtksourceview/stable/GtkSourceView.html
* /

/*
 * options
 * defined in sourceview.c
 *
 * -autoIdent
 * -highlightCurrentLine
 * -insertSpace
 * -marginsourceview
 * -showLineMarkers
 * -showLineNumbers
 * -smartHomeEnd
 * -tabsWidth
 */

#include <gnocl.h>

#include <string.h>
#include <assert.h>

#include <gtksourceview-2.0/gtksourceview/gtksourceview.h>
#include <gtksourceview-2.0/gtksourceview/gtksourcebuffer.h>
#include <gtksourceview-2.0/gtksourceview/gtksourcelanguage.h>
#include <gtksourceview-2.0/gtksourceview/gtksourcelanguagemanager.h>
#include <gtksourceview-2.0/gtksourceview/gtksourcestyleschememanager.h>
#include <gtksourceview-2.0/gtksourceview/gtksourceprintcompositor.h>

/*
#include <gtk/gtk.h>
#include <gtksourceview/gtksourceview.h>
#include <gtksourceview/gtksourcebuffer.h>
#include <gtksourceview/gtksourcelanguage.h>
#include <gtksourceview/gtksourcelanguagesmanager.h>
*/

Tcl_ObjCmdProc gnoclSourceViewCmd ;
int gnoclOptRightMargin ( Tcl_Interp *interp, GnoclOption *opt, GObject *obj, Tcl_Obj **ret );
int sourceViewFunc ( ClientData data, Tcl_Interp *interp, int objc, Tcl_Obj * const objv[] );
int gnoclSourceViewCommand ( GtkTextBuffer *buffer, Tcl_Interp *interp, int objc, Tcl_Obj * const objv[], int cmdNo, int isTextWidget );
int gnoclOptOnSourceViewUndo ( Tcl_Interp *interp, GnoclOption *opt, GObject *obj, Tcl_Obj **ret );
int gnoclOptOnSourceViewUndo ( Tcl_Interp *interp, GnoclOption *opt, GObject *obj, Tcl_Obj **ret );

static int open_file ( GtkSourceBuffer *sBuf, const gchar *filename );


/*---------------- formerly sourceview.c ----------------*/

static const int scrollBarIdx = 0;

/**
\brief
\author     William J Giddings
\since
**/
static void doOnUndoRedo  ( GtkSourceView *sourceview, gpointer user_data )
{
    GnoclCommandData *cs = ( GnoclCommandData * ) user_data;

    GnoclPercSubst ps[] =
    {
        { 'w', GNOCL_STRING },  /* widget */
        { 0 }
    };

    ps[0].val.str = gnoclGetNameFromWidget ( sourceview );
    gnoclPercentSubstAndEval ( cs->interp, ps, cs->command, 1 );
}

/**
\brief
\author     William J Giddings
\since
**/
int gnoclOptOnUndoRedo ( Tcl_Interp *interp, GnoclOption *opt, GObject *obj, Tcl_Obj **ret )
{
    assert ( *opt->propName == 'U' || *opt->propName == 'R' );
    return gnoclConnectOptCmd ( interp, obj, *opt->propName == 'U' ?  "undo" : "redo", G_CALLBACK ( doOnUndoRedo ), opt, NULL, ret );

}

/**
\brief
\author     William J Giddings
\since
**/
static int scrollToMark ( GtkTextView *view, GtkTextBuffer *buffer, Tcl_Interp *interp, int objc, Tcl_Obj * const objv[] )
{
#ifdef DEBUG_SOURCEVIEW
    printf ( "scrollToMark 1>  %s \n", Tcl_GetString ( objv[2] ) );
#endif

    GtkTextMark *mark;

    if ( objc < 3 )
    {
        Tcl_WrongNumArgs ( interp, 2, objv, "index ?-option val ...?" );
        return TCL_ERROR;
    }

    /* check to see that the mark exists */
    mark = gtk_text_buffer_get_mark ( buffer, Tcl_GetString ( objv[2] ) );

    /* reposition */
    if ( mark == NULL )
    {
        Tcl_SetResult ( interp, "This mark does not exist.", TCL_STATIC );
        return TCL_ERROR;
    }

    gtk_text_view_scroll_mark_onscreen  ( view, mark );

    return TCL_OK;

}

/**
\brief
\author
\date
**/
static int configure ( Tcl_Interp *interp, GtkScrolledWindow *scrolled, GtkTextView *text, GnoclOption options[] )
{
    if ( options[scrollBarIdx].status == GNOCL_STATUS_CHANGED )
    {
        GtkPolicyType hor, vert;

        if ( gnoclGetScrollbarPolicy ( interp, options[scrollBarIdx].val.obj, &hor, &vert ) != TCL_OK )
        {
            return TCL_ERROR;
        }

        gtk_scrolled_window_set_policy ( scrolled, hor, vert );
    }

    return TCL_OK;
}

/**
\brief
\author
\date
**/
int gnoclOptRightMargin ( Tcl_Interp *interp, GnoclOption *opt, GObject *obj, Tcl_Obj **ret )
{

    if ( 1 )
    {
        gtk_source_view_set_show_right_margin ( obj , 1 );

        return TCL_OK;
    }

    return TCL_ERROR;
}

static GnoclOption sourceViewOptions[] =
{
    /* sourceView properties */
    { "-autoIdent", GNOCL_BOOL, "auto-indent"},
    { "-drawSpaces", GNOCL_STRING, "draw-spaces" },
    { "-highlightCurrentLine", GNOCL_BOOL, "highlight-current-line" },
    { "-indentOnTab", GNOCL_BOOL, "indent-on-tab" },
    { "-indentWidth", GNOCL_INT, "indent-width" },
    { "-insertSpace", GNOCL_BOOL, "insert-spaces-instead-of-tabs" },
    { "-rightMargin", GNOCL_INT, "right-margin-position" },
    { "-showLineMarks", GNOCL_BOOL, "show-line-marks" },
    { "-showLineNumbers", GNOCL_BOOL, "show-line-numbers" },
    { "-showRightMargin", GNOCL_BOOL, "show-right-margin" },
    { "-smartHomeEnd", GNOCL_BOOL, "smart-home-end" },
    { "-tabWidth", GNOCL_INT, "tab-width"  },

    /* set these options through commands */
    { "-baseFont", GNOCL_OBJ, "Sans 14", gnoclOptGdkBaseFont },
    { "-baseColor", GNOCL_OBJ, "normal", gnoclOptGdkColorBase },

    /* textView events */
    { "-onButtonPress", GNOCL_OBJ, "P", gnoclOptOnButton },
    { "-onButtonRelease", GNOCL_OBJ, "R", gnoclOptOnButton },
    { "-onKeyPress", GNOCL_OBJ, "", gnoclOptOnKeyPress },
    { "-onKeyRelease", GNOCL_OBJ, "", gnoclOptOnKeyRelease },
    { "-onMotion", GNOCL_OBJ, "", gnoclOptOnMotion },

    /* sourceView signals */
    { "-onUndo", GNOCL_OBJ, "U", gnoclOptOnUndoRedo },
    { "-onRedo", GNOCL_OBJ, "R", gnoclOptOnUndoRedo },

    { NULL }
};

/**
\brief
\author
\date
**/
static int scrollToPos ( GtkTextView *view, GtkTextBuffer *buffer, Tcl_Interp *interp, int objc, Tcl_Obj * const objv[] )
{
#ifdef DEBUG_SOURCEVIEW
    printf ( "scrollToPos 1>  %s \n",  Tcl_GetString ( objv[2] ) );
#endif
    GnoclOption options[] =
    {
        { "-margin", GNOCL_DOUBLE, NULL },    /* 0 */
        { "-align", GNOCL_OBJ, NULL },        /* 1 */
        { NULL }
    };
    const int marginIdx = 0;
    const int alignIdx  = 1;

    int   ret = TCL_ERROR;

    double      margin = .0;
    int         useAlign = 0;
    gfloat      xAlign = 0.5, yAlign = 0.5;
    GtkTextIter iter;
    GtkTextMark *mark;


    if ( objc < 3 )
    {
        Tcl_WrongNumArgs ( interp, 2, objv, "index ?-option val ...?" );
        return TCL_ERROR;
    }

    if ( posToIter ( interp, objv[2], buffer, &iter ) != TCL_OK )
    {
        return TCL_ERROR;
    }

    if ( gnoclParseOptions ( interp, objc - 2, objv + 2, options ) != TCL_OK )
    {
        goto clearExit;
    }

    if ( options[alignIdx].status == GNOCL_STATUS_CHANGED )
    {
        if ( gnoclGetBothAlign ( interp, options[alignIdx].val.obj, &xAlign, &yAlign ) != TCL_OK )
        {
            goto clearExit;
        }

        useAlign = 1;
    }

    if ( options[marginIdx].status == GNOCL_STATUS_CHANGED )
    {
        margin = options[marginIdx].val.d;

        if ( margin < 0.0 || margin >= 0.5 )
        {
            Tcl_SetResult ( interp, "-margin must be between 0 and 0.5", TCL_STATIC );
            goto clearExit;
        }
    }


    mark = gtk_text_buffer_create_mark ( buffer, "__gnoclScrollMark__", &iter, 0 );


    gtk_text_view_scroll_to_mark ( view, mark, margin, useAlign, xAlign, yAlign );

    gtk_text_buffer_delete_mark ( buffer, mark );

    ret = TCL_OK;

clearExit:
    gnoclClearOptions ( options );
    return ret;
}


/**
\brief
\author
\date
**/
static int textInsert ( GtkTextBuffer *buffer, Tcl_Interp *interp, int objc, Tcl_Obj * const objv[], int cmdNo )
{
    GnoclOption insertOptions[] =
    {
        { "-tags", GNOCL_LIST, NULL },
        { NULL }
    };
    const int tagsIdx = 0;
    gint      startOffset;
    int       ret = TCL_ERROR;

    GtkTextIter   iter;

    if ( objc < cmdNo + 2 )
    {
        Tcl_WrongNumArgs ( interp, cmdNo, objv, "position text ?-option val ...?" );
        return TCL_ERROR;
    }


    if ( posToIter ( interp, objv[cmdNo], buffer, &iter ) != TCL_OK )
    {
        return TCL_ERROR;
    }

    if ( gnoclParseOptions ( interp, objc - cmdNo - 1, objv + cmdNo + 1, insertOptions ) != TCL_OK )
    {
        goto clearExit;
    }

    startOffset = gtk_text_iter_get_offset ( &iter );

    gtk_text_buffer_insert ( buffer, &iter, gnoclGetString ( objv[cmdNo+1] ), -1 );

    if ( insertOptions[tagsIdx].status == GNOCL_STATUS_CHANGED )
    {
        GtkTextIter start;
        int k, no;
        Tcl_Obj *obj = insertOptions[tagsIdx].val.obj;

        gtk_text_buffer_get_iter_at_offset ( buffer, &start, startOffset );

        if ( Tcl_ListObjLength ( interp, obj, &no ) != TCL_OK )
        {
            goto clearExit;
        }

        for ( k = 0; k < no; ++k )
        {
            Tcl_Obj *tp;

            if ( Tcl_ListObjIndex ( interp, obj, k, &tp ) != TCL_OK )
            {
                Tcl_SetResult ( interp, "Could not read tag list", TCL_STATIC );
                goto clearExit;
            }

            gtk_text_buffer_apply_tag_by_name ( buffer, Tcl_GetString ( tp ), &start, &iter );
        }
    }

    ret = TCL_OK;

clearExit:
    gnoclClearOptions ( insertOptions );

    return ret;
}

/**
\brief
\author
\date
**/
int sourceViewFunc ( ClientData data, Tcl_Interp *interp, int objc, Tcl_Obj * const objv[] )
{
    GtkScrolledWindow   *scrolled = GTK_SCROLLED_WINDOW ( data );
    GtkTextView         *text = GTK_TEXT_VIEW ( gtk_bin_get_child ( GTK_BIN ( scrolled ) ) );
    GtkTextBuffer       *buffer = gtk_text_view_get_buffer ( text );

    if ( objc < 2 )
    {
        Tcl_WrongNumArgs ( interp, 1, objv, "command" );
        return TCL_ERROR;
    }

    switch ( gnoclSourceViewCommand ( buffer, interp, objc, objv, 1, 1 ) )
    {
        case 0:
            break;  /* return TCL_OK */
        case 1:     /* delete */
            return gnoclDelete ( interp, GTK_WIDGET ( scrolled ), objc, objv );
        case 2:     /* configure */
            {
                int ret = TCL_ERROR;

                if ( gnoclParseAndSetOptions ( interp, objc - 1, objv + 1, sourceViewOptions, G_OBJECT ( text ) ) == TCL_OK )
                {
                    ret = configure ( interp, scrolled, text, sourceViewOptions );
                }

                gnoclClearOptions ( sourceViewOptions );

                return ret;
            }

            break;
        case 3: /* scrollToPosition */
            return scrollToPos ( text, buffer, interp, objc, objv );
        case 4: /* scrollToMark */
            return scrollToMark ( text, buffer, interp, objc, objv );
        default:
            return TCL_ERROR;
    }

    return TCL_OK;
}

/**
\brief
\author
\date
**/
int gnoclSourceViewCommand ( GtkTextBuffer *buffer, Tcl_Interp *interp, int objc, Tcl_Obj * const objv[], int cmdNo, int isTextWidget )
{
#ifdef DEBUG_SOURCEVIEW
    printf ( "sourceViewFunc\n" );
#endif
    const char *cmds[] = { "class", "delete", "configure", "scrollToPosition",  "scrollToMark",
                           "erase", "select", "get", "cut", "copy", "paste", "cget", "getLineCount", "getWordLength",
                           "getLength", "getLineLength", "setCursor", "getCursor", "insert", "tag", "dump", "XYgetCursor",
                           "signalEmit", "mark", "gotoWordStart", "gotoWordEnd", "search", "open",
                           NULL
                         };
    enum cmdIdx { ClassIdx, DeleteIdx, ConfigureIdx, ScrollToPosIdx, ScrollToMarkIdx,
                  EraseIdx, SelectIdx, GetIdx, CutIdx, CopyIdx, PasteIdx, CgetIdx, GetLineCountIdx, GetWordLengthIdx,
                  GetLengthIdx, GetLineLengthIdx, SetCursorIdx, GetCursorIdx, InsertIdx, TagIdx, DumpIdx, XYgetCursorIdx,
                  SignalEmitIdx, MarkIdx, GotoWordStartIdx, GotoWordEndIdx, SearchIdx, OpenIdx
                };

    int idx;

    if ( objc < 2 )
    {
        Tcl_WrongNumArgs ( interp, 1, objv, "command" );
        {
            return TCL_ERROR;
        }
    }

    if ( Tcl_GetIndexFromObj ( interp, objv[1], cmds, "command", TCL_EXACT, &idx ) != TCL_OK )
    {
        return TCL_ERROR;
    }


#ifdef DEBUG_SOURCEVIEW
    printf ( "0 = %s 1 = %s 2 = %s \n",
             Tcl_GetString ( objv[cmdNo+0] ) ,
             Tcl_GetString ( objv[cmdNo+1] ) ,
             Tcl_GetString ( objv[cmdNo+2] ) );
#endif

    switch ( idx )
    {
        case CutIdx:
        case CopyIdx:
        case PasteIdx:
            {
                /*  TODO: option which clipboard */
                GtkClipboard *clipboard = gtk_clipboard_get ( GDK_NONE );

                if ( objc != cmdNo + 1 )
                {
                    Tcl_WrongNumArgs ( interp, cmdNo + 1, objv, NULL );
                    return TCL_ERROR;
                }

                switch ( idx )
                {
                    case CutIdx:
                        gtk_text_buffer_cut_clipboard ( buffer, clipboard, 1 );
                        break;
                    case CopyIdx:
                        gtk_text_buffer_copy_clipboard ( buffer, clipboard );
                        break;
                    case PasteIdx:
                        gtk_text_buffer_paste_clipboard ( buffer, clipboard, NULL, 1 );
                        break;
                }
            }

            break;

        case EraseIdx:
        case SelectIdx:
        case GetIdx:
            {
#ifdef DEBUG_SOURCEVIEW
                printf ( "Erase/Select/Get -1\n" );
#endif
                GtkTextIter startIter, endIter;
                /*  text erase/select/getChars startIndex ?endIndex? */

                if ( objc < cmdNo + 2 || objc > cmdNo + 3 )
                {
                    Tcl_WrongNumArgs ( interp, cmdNo + 1, objv, "startIndex ?endIndex?" );
                    return TCL_ERROR;
                }

#ifdef DEBUG_SOURCEVIEW
                printf ( "Erase/Select/Get -2\n" );
#endif

                if ( posToIter ( interp, objv[cmdNo+1], buffer, &startIter ) != TCL_OK )
                {
#ifdef DEBUG_SOURCEVIEW
                    printf ( "Erase/Select/Get -2a\n" );
#endif

                    Tcl_WrongNumArgs ( interp, cmdNo + 1, objv, "startIndex ?endIndex?" );
                    return TCL_ERROR;
                }

#ifdef DEBUG_SOURCEVIEW
                printf ( "Erase/Select/Get -3\n" );
#endif

                if ( objc >= 4 )
                {
                    if ( posToIter ( interp, objv[cmdNo+2], buffer, &endIter ) != TCL_OK )
                    {
                        Tcl_WrongNumArgs ( interp, cmdNo + 1, objv, "startIndex ?endIndex?" );
                        return TCL_ERROR;
                    }
                }

                else
                {
                    endIter = startIter;
                    gtk_text_iter_backward_char ( &endIter );
                }

#ifdef DEBUG_SOURCEVIEW
                printf ( "Erase/Select/Get -4\n" );
#endif

                switch ( idx )
                {
                    case EraseIdx:
                        gtk_text_buffer_delete ( buffer, &startIter, &endIter );
                        break;
                    case SelectIdx:
                        gtk_text_buffer_place_cursor ( buffer, &startIter );
                        gtk_text_buffer_move_mark_by_name ( buffer, "selection_bound", &endIter );
                        break;
                    case GetIdx:
                        {
                            /*  TODO: include_hidden_chars */
                            char *txt = gtk_text_buffer_get_text ( buffer, &startIter, &endIter, 1 );
                            Tcl_SetObjResult ( interp, Tcl_NewStringObj ( txt, -1 ) );
                        }

                        break;
                }
            }

            break;

        case OpenIdx:
            {
#if DEBUG == 2
                g_print ( "objc= %d OpenIDx 0 = %s 1 = %s 2 = %s 3 = %s\n",
                          objc,
                          Tcl_GetString ( objv[cmdNo+0] ) ,
                          Tcl_GetString ( objv[cmdNo+1] ) ,
                          Tcl_GetString ( objv[cmdNo+2] ) ,
                          Tcl_GetString ( objv[cmdNo+3] ) );
#endif
                open_file ( buffer, Tcl_GetString ( objv[cmdNo+1] ) );
            }
            break;
        case ClassIdx:
            {
                Tcl_SetObjResult ( interp, Tcl_NewStringObj ( "text", -1 ) );
            }
            break;
        case SearchIdx:
            {
                int row1, col1, row2, col2;
                GtkTextIter start;
                GtkTextIter begin, end;
                Tcl_Obj *resList;
                gint applyTags;
                char *  pch;
                char *  tagList;

                if ( objc < cmdNo + 1 )
                {
                    Tcl_WrongNumArgs ( interp, cmdNo + 1, objv, "<search-phrase> -tags {taglist}" );
                    return -1;
                }

#ifdef DEBUG_SOURCEVIEW
                g_print ( "objc= %d SearchIDx 0 = %s 1 = %s 2 = %s 3 = %s\n",
                          objc,
                          Tcl_GetString ( objv[cmdNo+0] ) ,
                          Tcl_GetString ( objv[cmdNo+1] ) ,
                          Tcl_GetString ( objv[cmdNo+2] ) ,
                          Tcl_GetString ( objv[cmdNo+3] ) );
#endif

                if ( objc < cmdNo + 4 )
                {
                    if ( strcmp ( Tcl_GetString ( objv[cmdNo+2] ), "-tags" ) == 0 )
                    {
                        applyTags = 1;
                    }

                    else
                    {
                        applyTags = 0;
                    }
                }

                resList = Tcl_NewListObj ( 0, NULL );

                /*  default with the start of the buffer */
                gtk_text_buffer_get_start_iter ( buffer, &start );

                while ( gtk_source_iter_forward_search ( &start, Tcl_GetString ( objv[cmdNo+1] ), 0, &begin, &end, NULL ) != NULL )
                {

                    /*  return the index of the found location */
                    row1 = gtk_text_iter_get_line ( &begin );
                    col1 = gtk_text_iter_get_line_offset ( &begin );
                    row2 = gtk_text_iter_get_line ( &end );
                    col2 = gtk_text_iter_get_line_offset ( &end );

                    /*  check if there is a taglist to apply */
                    /*  PROBLEMS HERE... THE TAGLIST IS ONLY APPLIED TO THE FIRST ITEM, WHY? */

                    if ( applyTags )
                    {
                        pch = strtok ( Tcl_GetString ( objv[cmdNo+3] ), " " );

                        while ( pch != NULL )
                        {
                            g_print ( "apply tag = %s\n", pch );
                            gtk_text_buffer_apply_tag_by_name ( buffer, pch, &begin, &end );
                            pch = strtok ( NULL, " " );
                        }

                    }

                    Tcl_ListObjAppendElement ( interp, resList, Tcl_NewIntObj ( row1 ) );

                    Tcl_ListObjAppendElement ( interp, resList, Tcl_NewIntObj ( col1 ) );
                    Tcl_ListObjAppendElement ( interp, resList, Tcl_NewIntObj ( row2 ) );
                    Tcl_ListObjAppendElement ( interp, resList, Tcl_NewIntObj ( col2 ) );

                    start = end;

                }

                Tcl_SetObjResult ( interp, resList );

                return TCL_OK;
            }


            break;
        case InsertIdx:

            /*
             * GET A DIRECT INPUT OF TEXT...
             */

            // gtk_text_buffer_set_text (GTK_TEXT_BUFFER( buffer ), "Your 1st GtkSourceView widget!", -1);

            if ( textInsert ( buffer, interp, objc, objv, cmdNo + 1 ) != TCL_OK )
            {
                return -1;
            }

            /* this action needs to emit an insert text signal */

            break;

        case TagIdx:
            {
                if ( tagCmd ( buffer, interp, objc, objv, cmdNo + 1 ) != TCL_OK )
                {
                    return -1;
                }
            }
            break;

        default:
            assert ( 0 );

            return -1;
    }

    return TCL_OK;
}

/**
\brief
\author
\date
**/
int gnoclSourceViewCmd ( ClientData data, Tcl_Interp *interp, int objc, Tcl_Obj * const objv[]  )
{
    int               ret;
    GtkWidget*  sourceView;
    GtkScrolledWindow *scrolled;

#ifdef DEBUG_SOURCEVIEW
    printf ( "gnoclSourceViewCmd -1\n" );
#endif

    if ( gnoclParseOptions ( interp, objc, objv, sourceViewOptions ) != TCL_OK )
    {
        gnoclClearOptions ( sourceViewOptions );
        return TCL_ERROR;
    }

    sourceView = gtk_source_view_new ();

    if ( 0 )
    {
        gtk_source_view_set_show_line_numbers ( sourceView, TRUE );
        gtk_source_view_set_highlight_current_line ( sourceView, TRUE );
    }

    scrolled =  GTK_SCROLLED_WINDOW ( gtk_scrolled_window_new ( NULL, NULL ) );

    gtk_scrolled_window_set_policy ( scrolled, GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC );
    gtk_container_add ( GTK_CONTAINER ( scrolled ), GTK_WIDGET ( sourceView ) );
    gtk_widget_show_all ( GTK_WIDGET ( scrolled ) );

    ret = gnoclSetOptions ( interp, sourceViewOptions, G_OBJECT ( sourceView ), -1 );

    if ( ret == TCL_OK )
    {
        ret = configure ( interp, scrolled, sourceView, sourceViewOptions );
    }

    gnoclClearOptions ( sourceViewOptions );


    if ( ret != TCL_OK )
    {
        gtk_widget_destroy ( GTK_WIDGET ( scrolled ) );
        return TCL_ERROR;
    }

    return gnoclRegisterWidget ( interp, GTK_WIDGET ( scrolled ), sourceViewFunc );
}

/**
\brief      Initialize the gnoc::sourceview widget
\author     William J Giddings
\note       The Gnocl initilisation proceedure. It is necessary to end the name of this function
            with "_Init" in order to initialise a shared library. This function will be automatically
            called when the Tcl load command loads the library.
\see        Welch, Jones & Hobbs (2003) pp695-699
**/
int Gnoclsourceview_Init ( Tcl_Interp *interp )
{
#ifdef DEBUG_SOURCEVIEW
    printf ( "Initializing gnocl sourceView version %s\n", VERSION );
#endif

    if ( Tcl_InitStubs ( interp, "8.3", 0 ) == NULL )
        return TCL_ERROR;

    if ( Tcl_PkgRequire ( interp, "Gnocl", VERSION, 0 ) == NULL )
        return TCL_ERROR;

    if ( Tcl_PkgProvide ( interp, "GnoclSourceView", VERSION ) != TCL_OK )
        return TCL_ERROR;

    /* gnoclSourceViewCmd is located in sourceview.c */
    Tcl_CreateObjCommand ( interp, "gnocl::sourceView", gnoclSourceViewCmd, NULL, NULL );

    return TCL_OK;
}

/**
\brief
\author
\see        http://www.bravegnu.org/gtktext/x561.html#AEN566
**/
static int open_file ( GtkSourceBuffer *sBuf, const gchar *filename )
{

#ifdef DEBUG_SOURCEVIEW
    g_print ( "open_file %s\n", filename );
#endif

    return TCL_OK;


    GtkSourceLanguageManager *lm;
    GtkSourceLanguage *language = NULL;
    GError *err = NULL;
    gboolean reading;
    GtkTextIter iter;
    GIOChannel *io;
    gchar *buffer;

    g_return_val_if_fail ( sBuf != NULL, FALSE );
    g_return_val_if_fail ( filename != NULL, FALSE );
    g_return_val_if_fail ( GTK_IS_SOURCE_BUFFER ( sBuf ), FALSE );

    /* get the Language for C source mimetype */
    lm = g_object_get_data ( G_OBJECT ( sBuf ), "languages-manager" );

    language = gtk_source_language_manager_guess_language ( lm, filename, NULL );
    g_print ( "Language: [%s]\n", gtk_source_language_get_name ( language ) );

    if ( language == NULL )
    {
        g_print ( "No language found for mime type `%s'\n", "text/x-csrc" );
        g_object_set ( G_OBJECT ( sBuf ), "highlight", FALSE, NULL );
    }

    else
    {
        gtk_source_buffer_set_language ( sBuf, language );
        g_object_set ( G_OBJECT ( sBuf ), "highlight", TRUE, NULL );
    }

    /* Now load the file from Disk */
    io = g_io_channel_new_file ( filename, "r", &err );

    if ( !io )
    {
        g_print ( "error: %s %s\n", ( err )->message, filename );
        return FALSE;
    }

    if ( g_io_channel_set_encoding ( io, "utf-8", &err ) != G_IO_STATUS_NORMAL )
    {
        g_print ( "err: Failed to set encoding:\n%s\n%s", filename, ( err )->message );
        return FALSE;
    }

    gtk_source_buffer_begin_not_undoable_action ( sBuf );

    //gtk_text_buffer_set_text (GTK_TEXT_BUFFER (sBuf), "", 0);
    buffer = g_malloc ( 4096 );
    reading = TRUE;

    while ( reading )
    {
        gsize bytes_read;
        GIOStatus status;

        status = g_io_channel_read_chars ( io, buffer, 4096, &bytes_read, &err );

        switch ( status )
        {
            case G_IO_STATUS_EOF: reading = FALSE;

            case G_IO_STATUS_NORMAL:

                if ( bytes_read == 0 ) continue;

                gtk_text_buffer_get_end_iter ( GTK_TEXT_BUFFER ( sBuf ), &iter );
                gtk_text_buffer_insert ( GTK_TEXT_BUFFER ( sBuf ), &iter, buffer, bytes_read );
                break;

            case G_IO_STATUS_AGAIN: continue;

            case G_IO_STATUS_ERROR:

            default:
                g_print ( "err (%s): %s", filename, ( err )->message );
                /* because of error in input we clear already loaded text */
                gtk_text_buffer_set_text ( GTK_TEXT_BUFFER ( sBuf ), "", 0 );

                reading = FALSE;
                break;
        }
    }

    g_free ( buffer );

    gtk_source_buffer_end_not_undoable_action ( sBuf );
    g_io_channel_unref ( io );

    if ( err )
    {
        g_error_free ( err );
        return FALSE;
    }

    gtk_text_buffer_set_modified ( GTK_TEXT_BUFFER ( sBuf ), FALSE );

    /* move cursor to the beginning */
    gtk_text_buffer_get_start_iter ( GTK_TEXT_BUFFER ( sBuf ), &iter );
    gtk_text_buffer_place_cursor ( GTK_TEXT_BUFFER ( sBuf ), &iter );

    g_object_set_data_full ( G_OBJECT ( sBuf ), "filename", g_strdup ( filename ),
                             ( GDestroyNotify ) g_free );

    return TRUE;
}
